#!/usr/bin/python
#
# NAME: compute_mean_annual_cycle.py
#
# PURPOSE: Reads daily mean SM and ST time series data files and generates mean and annual
#          cycle data
#
# INPUTS:  NAME      type       dimension       description
#          --------------------------------------------------------------------------
#          end_year  long       scalar          last year of climatology
#
# OUTPUTS:
#          SM_tseries_mean.txt       -> ASCII file containing long term SM mean for each station
#          SM_tseries_ac_leapday.txt -> ASCII file containing SM annual cycle for each station
#                                       on leap years
#          SM_tseries_ac_noleapday.txt -> ASCII file containing SM annual cycle for each station on
#                                         non-leap years
#          ST*.txt -> Same as SM data only for soil temperature data
#
# NOTES:
#          Only complete calendar years are used from input time series.
#
# CREATION HISTORY:  Darren Jackson CIRES/PSL September 2021

import sys
import pandas as pd
import numpy as np

# Read SM and ST time series data

dpath = '../data/training_data/'
fileName_SM = dpath+'SM_tseries.txt'
fileName_ST = dpath+'ST_tseries.txt'
smdata0 = pd.read_csv(fileName_SM,sep='\s+') #/s+ is whitespace separater
stdata0 = pd.read_csv(fileName_ST,sep='\s+') #/s+ is whitespace separater

# Remove any partial years of data at start and/or end of each time series

smdata0['month'] = pd.DatetimeIndex(smdata0['Date']).month
smdata0['day'] = pd.DatetimeIndex(smdata0['Date']).day
month = pd.DatetimeIndex(smdata0['Date']).month
day = pd.DatetimeIndex(smdata0['Date']).day
start_years = ((smdata0['month'] == 1) & (smdata0['day'] == 1))
end_years = ((smdata0['month'] == 12) & (smdata0['day'] == 31))
smdata1 = smdata0.loc[start_years]
smdata2 = smdata0.loc[end_years]
start_date = smdata1.iloc[0]['Date']
end_date = smdata2.iloc[-1]['Date']

stdata0['month'] = pd.DatetimeIndex(stdata0['Date']).month
stdata0['day'] = pd.DatetimeIndex(stdata0['Date']).day
start_years = ((stdata0['month'] == 1) & (stdata0['day'] == 1))
end_years = ((stdata0['month'] == 12) & (stdata0['day'] == 31))
stdata1 = stdata0.loc[start_years]
stdata2 = stdata0.loc[end_years]
start_date_t = stdata1.iloc[0]['Date']
end_date_t = stdata2.iloc[-1]['Date']
if start_date != start_date_t or end_date != end_date_t:
    sys.exit('SM and ST data have incompatible start and end dates, exit compute_mean_annual_cycle.py program')
    
smdata = smdata0[(smdata0['Date'] >= start_date) & (smdata0['Date'] <= end_date)]
stdata = stdata0[(stdata0['Date'] >= start_date_t) & (stdata0['Date'] <= end_date_t)]

# Compute total mean for each station

sm_only = smdata.iloc[:,1:-2]  # -2 eliminates month and day columns added previously
st_only = stdata.iloc[:,1:-2]

sm_means = sm_only.mean(axis = 0, skipna = True)
st_means = st_only.mean(axis = 0, skipna = True)

# Center sm and st time series on zero by subtracting out the mean

sm_only_Array = sm_only.to_numpy()
st_only_Array = st_only.to_numpy()

n_times, n_stations = sm_only.shape
sm_cent = np.zeros((n_times,n_stations))
st_cent = np.zeros((n_times,n_stations))
for n in np.arange(0, n_times):
    sm_cent[n,:] = sm_only_Array[n,:]-sm_means[:]
    st_cent[n,:] = st_only_Array[n,:]-st_means[:]

# Create 2nd set of centered arrays with leap year days (Feb 29th) removed

noleapdays = ~((smdata['month'] == 2) & (smdata['day'] == 29))
sm_cent_noleapday = sm_cent[noleapdays,:]
st_cent_noleapday = st_cent[noleapdays,:]

# Find start and end year of input data

smdata['year'] = pd.DatetimeIndex(smdata['Date']).year
start_year = smdata['year'].min()
end_year = smdata['year'].max()
yrList = np.arange(start_year,end_year+1)
n_years = len(yrList)
yrStarts = np.empty([n_years])

# Find index for the first day of each year for no leap day time series

smdata['leap'] = smdata['Date'].str.find('-02-29')
noleapdays = ~(smdata['leap'] != -1)
smdata_noleap = smdata[noleapdays]
date_noleap = smdata_noleap.iloc[:,0]
date_noleap_Array = date_noleap.to_numpy()

for n in np.arange(0, n_years):
    yrStarts[n] = np.where(date_noleap_Array == str(yrList[n])+'-01-01')[0]
    

# Change arrays from station x total_days to station x year x day_of_year

yearlySM = np.full([n_years,365,n_stations],np.nan)
yearlyST = np.full([n_years,365,n_stations],np.nan)

for iYr in np.arange(0,n_years):
    if iYr==0:  # If first year of data
        if yrStarts[0]==0:
            yearlySM[iYr,0:int(yrStarts[1]),:] = sm_cent_noleapday[0:int(yrStarts[1]),:]
            yearlyST[iYr,0:int(yrStarts[1]),:] = st_cent_noleapday[0:int(yrStarts[1]),:]
        else:
            yearlySM[iYr,int(yrStarts[0])::,:] = sm_cent_noleapday[0:int(yrStarts[0]),:]
            yearlyST[iYr,int(yrStarts[0])::,:] = st_cent_noleapday[0:int(yrStarts[0]),:]        
    elif iYr==n_years-1:   # If last year of data
        if yrStarts[0]==0:
            yearlySM[iYr,:,:] = sm_cent_noleapday[int(yrStarts[iYr]):n_times,:]
            yearlyST[iYr,:,:] = st_cent_noleapday[int(yrStarts[iYr]):n_times,:]
        else:
            yearlySM[iYr,:,:] = sm_cent_noleapday[int(yrStarts[iYr-1]):n_times,:]
            yearlyST[iYr,:,:] = st_cent_noleapday[int(yrStarts[iYr-1]):n_times,:] 
    else:                    # all other years
        if yrStarts[0]==0:
            yearlySM[iYr,:,:] = sm_cent_noleapday[int(yrStarts[iYr]):int(yrStarts[iYr+1]),:]
            yearlyST[iYr,:,:] = st_cent_noleapday[int(yrStarts[iYr]):int(yrStarts[iYr+1]),:]
        else:
            yearlySM[iYr,:,:] = sm_cent_noleapday[int(yrStarts[iYr-1]):int(yrStarts[iYr]),:]
            yearlyST[iYr,:,:] = st_cent_noleapday[int(yrStarts[iYr-1]):int(yrStarts[iYr]),:] 

# annual average cycle
                    
# Non-leap years, stations x 365
annCycle_SM_no_leap = np.nanmean(yearlySM,axis=0)
annCycle_ST_no_leap = np.nanmean(yearlyST,axis=0)

# Leap years, stations x 366
annCycle_SM_leap = np.full([366,n_stations],np.nan)
annCycle_ST_leap = np.full([366,n_stations],np.nan)

annCycle_SM_leap[0:60,:]=annCycle_SM_no_leap[0:60,:]
annCycle_ST_leap[0:60,:]=annCycle_ST_no_leap[0:60,:]
tmp_SM = annCycle_SM_no_leap[59:61,:]
tmp_ST = annCycle_ST_no_leap[59:61,:]
annCycle_SM_leap[60,:]=np.nanmean(tmp_SM,axis=0)
annCycle_ST_leap[60,:]=np.nanmean(tmp_ST,axis=0)
annCycle_SM_leap[61:366,:]=annCycle_SM_no_leap[60:365,:]
annCycle_ST_leap[61:366,:]=annCycle_ST_no_leap[60:365,:]

# Write mean and annual cycle data to files
# Write daily mean soil moisture and temperature time series for all stations
# Note that time unit is not included with annual cycle data
# noleapday files are 365 days and leapday files are 366 days

np.savetxt(dpath+"SM_tseries_mean.txt",[sm_means], fmt='%7.4f')
np.savetxt(dpath+"SM_tseries_ac_noleapday.txt",annCycle_SM_no_leap, fmt='%7.4f')
np.savetxt(dpath+"SM_tseries_ac_leapday.txt",annCycle_SM_leap, fmt='%7.4f')

np.savetxt(dpath+"ST_tseries_mean.txt",[st_means], fmt='%7.3f')
np.savetxt(dpath+"ST_tseries_ac_noleapday.txt",annCycle_ST_no_leap, fmt='%7.3f')
np.savetxt(dpath+"ST_tseries_ac_leapday.txt",annCycle_ST_leap, fmt='%7.3f')

