#!/usr/bin/python3
#
# NAME: convert_forecasted_zscores_to_vwc.py
#
# PURPOSE: Reads soil moisture and temperature forecasted
#          z-scores derived from soil_moisture_forecast.py
#          and generate soil moisture in volumetric water content units
#
#
# INPUT DATA:
#
#          SM_initial_condition.txt     -> Reads soil moisture input file to retrieve date information
#          forecasted_SM_zscores.p      -> forecasted SM z-scores needed to convert to VWC
#
#
# OUTPUT DATA:
#
#          SM_<yyyymmdd>_forecast.csv   -> SM forecast data from LIM for initial condition at date yyyymmdd
#
# NOTES:
#        Use same format for IC data as original soil moisture and
#        and temperature data used to train the LIM.
#
# CREATION HISTORY:  Darren Jackson, CIRES/ESRL/PSL, October 2021
#----------------------------------------------------------------------

import numpy as np 
import os, sys
import pandas as pd
import pickle
import csv

# Read initial condition date
yyyymmdd = sys.argv[1]

#-------------------------------------------
# Read forecasted zscores to output file
#-------------------------------------------

fpath = '../data/forecast_data/'
ipath = fpath+'initial_conditions/'
tpath = '../data/training_data/'

# Read initial condition to retrieve dates

fileName_SM = ipath+'SM_'+yyyymmdd+'_ic.txt'
smdata = pd.read_csv(fileName_SM,sep='\s+') #/s+ is whitespace separater
stationlist = list(smdata.columns)
stationlist.insert(0,'Lead')

# Retrieve forecasted SM z-scores

f = open(fpath+"Zscores_"+yyyymmdd+"_forecast.p","rb")
FoO = pickle.load(f)
cutoff_F = pickle.load(f)
fcst_zscores = pickle.load(f)
Exp_stationErr = pickle.load(f)
f.close()

# Read SM and ST mean and annual cycle data

sm_means = np.loadtxt(tpath+"SM_tseries_mean.txt")
annCycle_SM_no_leap = np.loadtxt(tpath+"SM_tseries_ac_noleapday.txt")
annCycle_SM_leap = np.loadtxt(tpath+"SM_tseries_ac_leapday.txt")

st_means = np.loadtxt(tpath+"ST_tseries_mean.txt")
annCycle_ST_no_leap = np.loadtxt(tpath+"ST_tseries_ac_noleapday.txt")
annCycle_ST_leap = np.loadtxt(tpath+"ST_tseries_ac_leapday.txt")

# Read the means and std data used to compute z-scores
f = open(tpath+"anomaly_mean_std_regr_stats.p","rb")
allArray_mean = pickle.load(f)
allArray_std = pickle.load(f)
regr = pickle.load(f)
f.close()

#---------------------------------------------------------------
# Compute soil moisture anomalies from z-score forecast data
#---------------------------------------------------------------

sizes = np.shape(fcst_zscores)
nDat = sizes[0]
nT = sizes[1]
nlags = sizes[2]
nDat2 = int(nDat/2)

#-----------------------------------------------------------------------------
# Add back in the z-score std and mean removed to normalized the anomaly data
#-----------------------------------------------------------------------------

sm_ac_removed = np.full([nT,nDat2,nlags], np.nan) # note I switched time and station (timexstationxlag)
err_ac_removed = np.full([nT,nDat2,nlags], np.nan)
for iSt in np.arange(nDat2):
    for iT in np.arange(nT):
        sm_ac_removed[iT,iSt,:] = (fcst_zscores[iSt+nDat2,iT,:]*allArray_std[iSt+nDat2])+allArray_mean[iSt+nDat2]
        err_ac_removed[iT,iSt,:] = ((fcst_zscores[iSt+nDat2,iT,:]+Exp_stationErr[:,iSt+nDat2])*allArray_std[iSt+nDat2])+allArray_mean[iSt+nDat2]
        
#-----------------------------------------------------------------
# Add the SM annual cycle and total mean back to the anomaly field
#-----------------------------------------------------------------

smdata['doy'] = pd.DatetimeIndex(smdata['Date']).dayofyear
smdata['leap'] = pd.DatetimeIndex(smdata['Date']).is_leap_year

# Define SM array defined in VWC
sm_vwc = np.full([nT,nDat2,nlags],np.nan)
err_vwc = np.full([nT,nDat2,nlags],np.nan)

for n in np.arange(0, nT):
    m = smdata['doy'][n] - 1
    for ilag in np.arange(0,nlags):
        if smdata['leap'][n]:
            sm_vwc[n,:,ilag] = sm_ac_removed[n,:,ilag] + annCycle_SM_leap[m,:] + sm_means[:]
            err_vwc[n,:,ilag]  = err_ac_removed[n,:,ilag] + annCycle_SM_leap[m,:] + sm_means[:]
        else:
            sm_vwc[n,:,ilag] = sm_ac_removed[n,:,ilag] + annCycle_SM_no_leap[m,:] + sm_means[:]
            err_vwc[n,:,ilag]  = err_ac_removed[n,:,ilag] + annCycle_SM_no_leap[m,:] + sm_means[:]

#----------------------------------------------------
# Write soil moisture VWC forecast to output file
#----------------------------------------------------

fileName_SM = '../data/forecast_data/SM_'+yyyymmdd+'_forecast.csv'
fileName_ER = '../data/forecast_data/SM_'+yyyymmdd+'_forecast_error.csv'

stationlist.insert(0,'FoO') 

with open(fileName_SM,'w+') as f:
    writer = csv.writer(f)

 # Write header 
    writer.writerow(stationlist)

 # Write forecast data. Only write 21 days
    smdata['Date'] = pd.to_datetime(smdata['Date'],format='%Y-%m-%d')
    for ilag in range(22):
        smdata['Date_new'] = smdata['Date']+pd.to_timedelta(ilag,unit='d')
        sm_list = np.round(sm_vwc[0,:,ilag],4).tolist()
        sm_list.insert(0,smdata['Date_new'][0].strftime('%Y-%m-%d'))
        sm_list.insert(0,ilag)
        if FoO == 1:
            sm_list.insert(0,1)
        else:
            sm_list.insert(0,0)
            
        writer.writerow(sm_list)

with open(fileName_ER,'w+') as f:
    writer = csv.writer(f)

 # Write header
    writer.writerow(stationlist)

 # Write forecast error data
    smdata['Date'] = pd.to_datetime(smdata['Date'],format='%Y-%m-%d')
    for ilag in range(22):
        smdata['Date_new'] = smdata['Date']+pd.to_timedelta(ilag,unit='d')
        err_list = np.round(err_vwc[0,:,ilag]-sm_vwc[0,:,ilag],4).tolist()
        err_list.insert(0,smdata['Date_new'][0].strftime('%Y-%m-%d'))
        err_list.insert(0,ilag)
        if FoO == 1:
            err_list.insert(0,1)
        else:
            err_list.insert(0,0)

        writer.writerow(err_list)
