#!/usr/bin/python3
#
# NAME: error_analysis.py
#
# PURPOSE: Computes hindcast errors from LIM parameters generated from the training data and
#          determines if forecast of opportunity are skillful
#
# INPUT DATA: Name                       Type     Dimension           Origin Code
#             -----------------------------------------------------------------------------
#             zScores_residST_SM.p       Pickle   [stations x time]   compute_anomaly_zscores_residuals.py
#             optimal_structure_peak.p   Pickle   Tau_peak - Integer        optimal_structure.py
#                                                 gamma_all - [time]        optimal_structure.py
#                                                 optimStruct - [stations]  optimal_structure.py
#             optimal_projections.p      Pickle   projTrcile - float        optimal_structure.py
#                                                 optProj - [stations]   
#             lim_forecast_matrics.p     Pickle   [leads x stations x stations]    train_lim.py
#             optimal_projections_pcor.p Pickle   pcor_Trcile - float       optimal_structure.py
#                                                 pcorr - [stations]  
# OUTPUT:
#             To the screen, outputs if forecast of opportunities can be identified using
#             inputted LIM parameters.  If so, then outputs if Projection or Correlation
#             method is best method
#
# CREATION HISTORY: Darren Jackson, CIRES/PSL, December 2021
#------------------------------------------------------------------------------------------

import numpy as np 
import os, sys
import pandas as pd
import pickle 

#------------------------
# Read Input Data
#------------------------

ipath = '../data/training_data/'
dpath = '../data/lim_parameter_data/'
# Input zScores, forecast matrices, optimal structure, and projection data
filePath_zscore = ipath+'/zScores_residST_SM.p'
if os.path.isfile(filePath_zscore)==True:
    data = pickle.load( open( filePath_zscore, "rb" ) )
else: 
    sys.exit('   WARNING: Zscore data file does not exist! Cannot continue.')
    
filePath_structure = dpath+'optimal_structure_peak.p'
if os.path.isfile(filePath_structure)==True:
    f = open(filePath_structure,"rb")
    Tau_peak = pickle.load(f)
    optimStruct = pickle.load(f)
    f.close()
else:
     sys.exit('   WARNING: Optimal structure data file does not exist! Cannot continue.')
            
filePath_forecast_matrices =dpath+'lim_forecast_matrices.p'
if os.path.isfile(filePath_forecast_matrices)==True:
    f = open(filePath_forecast_matrices,"rb")
    tau0 = pickle.load(f)
    G_tau_3d = pickle.load(f)
    Exp_StationErr = pickle.load(f)
    f.close()
else: 
    sys.exit('   WARNING: forecast matrix data file does not exist! Cannot continue.')

filePath_pcorr = dpath+'optimal_projection_pcorr.p'
if os.path.isfile(filePath_pcorr)==True:
    f = open(filePath_pcorr,"rb")
    pcorr_Trcile = pickle.load(f)
    pcorr = pickle.load(f)
    f.close()
else:
    sys.exit('   WARNING: Correlation data file does not exist! Cannot continue.')

print('Lag time in days of best forecast skill: ',tau0)
print('Lag time in days of maximum amplification: ',Tau_peak)
print('Top tercile cut-off value: ',pcorr_Trcile)

# Optimal structure using projection method turned off
#filePath_projection = dpath+'optimal_projection.p'
#if os.path.isfile(filePath_projection)==True:
#    f = open(filePath_projection,"rb")
#    projTrcile = pickle.load(f)
#    optProj = pickle.load(f)
#    f.close()
#else:
#    sys.exit('   WARNING: Projection histogram data file does not exist! Cannot continue.')

#------------------------------------------
# Define parameters need to assign arrays
#------------------------------------------

# Define size of data array 
sizes = np.shape(data)
nDat  = sizes[0]
nT    = sizes[1]
nDat2 = int(nDat/2)

# Number of lags determined in train_lim.ipynb G_tau output data
Gsize = np.shape(G_tau_3d)
n_lags = Gsize[0]

#-------------------------------------
# Determine days in top tercile 
#-------------------------------------

#Find days where the value of the projection is above the cutoff (in the top tercile)
#iProj = np.where(np.abs(optProj)>=projTrcile)
#topDays = iProj[0]

#Find days where the value of the pcorr projection is above the cutoff (in the top tercile)
iProj = np.where(np.abs(pcorr)>=pcorr_Trcile)
topDays_pcorr = iProj[0]

#------------------------------------------------------------
# Determine error using optimal structure with tercile data
#------------------------------------------------------------

#   Create empty arrays to store values in 
meanErr_fcst         = np.full([nDat, n_lags], np.nan) 
meanErr_fcst_pcorr_optDays = np.full([nDat, n_lags], np.nan) 
#meanErr_fcst_optDays = np.full([nDat, n_lags], np.nan) 

for ilag in range(1,n_lags):   #Loop over number of lags to be computed. Error at ilag=0 set to zero later.
    # Set up initial values as zero for time-averaged values 
    eps2_L = np.full([nDat],0.)
    ncor_L = np.full([nDat],0.)

    eps2_optL = np.full([nDat],0.)
    ncor_optL = np.full([nDat],0.)

    eps2_pcorr_optL = np.full([nDat],0.)
    ncor_pcorr_optL = np.full([nDat],0.)

    #Define Green function at this lag using input data
    G_tau = G_tau_3d[ilag,:,:]
    
    #Define empty array to store forecast at each time
    fcst     = np.full([nDat,nT-ilag], np.nan)
    
    for iT in range(nT-ilag):    #Loop over timesteps 
        
        # --- Get forecast data --- #
        #If there is any missing data at this time...
        if ( np.isnan(data[:,iT]).any() ): 
            #Indices where NaNs are *not*
            iNan = np.where(~np.isnan(data[:,iT]))
            
            #Make replica matrix replacing NaNs with zeros so that it won't count for other stations (but is computed)
            replZeros          = np.zeros(np.shape(data[:,iT]))
            replZeros[iNan[0]] = data[iNan[0],iT]
            
            #Get forecast using zeros array so that stations w/o NaN are still computed properly 
            fcst[:,iT] = np.dot(G_tau, replZeros)
            
            #Replacing stations that had missing data with a NaN (no forecast if no data)
            fcst[np.where(np.isnan(data[:,iT])),iT] = np.nan
        else:     
            #If no data is missing, just compute forecast as normal - G(Tau) * X(:,t)
            fcst[:,iT] = np.dot(G_tau, data[:,iT])
            
        # -- Build up mean squared error estimates for the LIM forecast at each station -- #
        for iSt in range(nDat):     #Loop over stations 
            # Make sure data exists
            if ( np.isfinite( data[iSt,iT] ) and np.isfinite( data[iSt,iT+ilag] ) ):     
                fcst_Err    = data[iSt,iT+ilag] - fcst[iSt,iT]
                eps2_L[iSt] = np.nansum([fcst_Err**2, eps2_L[iSt]])
                ncor_L[iSt] = ncor_L[iSt] + 1
                
#                if ( (iT+ilag) in topDays ):
#                    eps2_optL[iSt] = np.nansum([fcst_Err**2, eps2_optL[iSt]])
#                    ncor_optL[iSt] = ncor_optL[iSt] + 1
                    
                if( (iT+ilag) in topDays_pcorr ):
                    eps2_pcorr_optL[iSt] = np.nansum([fcst_Err**2, eps2_pcorr_optL[iSt]])
                    ncor_pcorr_optL[iSt] = ncor_pcorr_optL[iSt] + 1

    # Compute mean squared error for every station over the period at this lag 
    meanErr_fcst[:,ilag]         = eps2_L/ncor_L 
    meanErr_fcst_pcorr_optDays[:,ilag] = eps2_pcorr_optL/ncor_pcorr_optL
#    meanErr_fcst_optDays[:,ilag] = eps2_optL/ncor_optL
    
# Set first error estimate to zero: 
meanErr_fcst[:,0]         = 0
meanErr_fcst_pcorr_optDays[:,0] = 0
#meanErr_fcst_optDays[:,0] = 0

#Normalize by the variance at each station 
for iSt in range(nDat):
    meanErr_fcst[iSt,:]         = meanErr_fcst[iSt,:]/np.nanvar(data[iSt,:])
    meanErr_fcst_pcorr_optDays[iSt,:] = meanErr_fcst_pcorr_optDays[iSt,:]/np.nanvar(data[iSt, topDays_pcorr])
#    meanErr_fcst_optDays[iSt,:] = meanErr_fcst_optDays[iSt,:]/np.nanvar(data[iSt, topDays])

#---------------------------------------------------------------
# Compute error averaged over all stations and 7-14 day period
#---------------------------------------------------------------

#   Find mean error in 7-14 day lead time for all soil moisture station data
meanErr_fcst_all = np.nanmean(meanErr_fcst[int(nDat/2):nDat-1,7:15])
meanErr_fcst_pcorr_optDays_all = np.nanmean(meanErr_fcst_pcorr_optDays[int(nDat/2):nDat-1,7:15])
#meanErr_fcst_optDays_all = np.nanmean(meanErr_fcst_optDays[int(nDat/2):nDat-1,7:15])

print('Mean forecast error = ',meanErr_fcst_all)
print('Mean error for forecasts of opportunity using Correlation = ',meanErr_fcst_pcorr_optDays_all)

if (meanErr_fcst_all < meanErr_fcst_pcorr_optDays_all):
    print('Do not use forecast model. Forecast of Opportunity error > Total mean forecast error')
else:
    print('Forecast model is ready. Forecast of Opportunity error < Total mean forecast error')

