#!/usr/bin/python3
#
# NAME: optimal_structure.py
#
# PURPOSE: Use Green functions from train_lim.py program to determine the lead
#          corresponding to the maximum amount of growth and the corresponding
#          optimal initail condition pattern
#
# INPUT FILES:
#   Name                    Type     Dimension                     Origin Code
#   ---------------------------------------------------------------------------------------
#   zScores_residST_SM.p    Pickle   [stations x time]             compute_anomaly_zscores_residuals.py
#   lim_forecast_matrices.p Pickle   [leads x stations x stations] train_lim.py
#
# OUTPUT FILE:
#              optimal_structure_tercile.p -> Contains lag time of maximum growth, the optimal growth
#                                             structure, and minimum correlation needed between optimal
#                                             structure and initial conditions to define an optimal
#                                             forecast
#                                             
# NOTES:
#
# CREATION HISTORY:  Meg D. Fowler, CIRES/PSL, February 2020
#                    Modified by Darren Jackson, CIRES/PSL, October 2021
#
# Import Libraries 

import numpy as np 
from   numpy import linalg as LA
import scipy 
from   scipy import stats
import pandas as pd
import pickle
import os, sys

# Function written so that all of LIM is self-contained/can be called by a single command
from LIM_analysis import LIM   

#------------------------------------------------------------------
# Define file paths to zscore data and lim forecast matrices
#------------------------------------------------------------------
ipath = '../data/training_data/'
dpath = '../data/lim_parameter_data/'
filePath_zscore = ipath+'zScores_residST_SM.p'
filePath_forecast_matrices =dpath+'lim_forecast_matrices.p'

#-----------------------------------------------------
# Load in pickle files containing user-supplied data 
#-----------------------------------------------------

if os.path.isfile(filePath_zscore)==True:
    data = pickle.load( open( filePath_zscore, "rb" ) )
else: 
    print('!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--')
    print('   WARNING: Zscore data file does not exist! Cannot continue.')
    print('!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--')
    sys.exit()
    
if os.path.isfile(filePath_forecast_matrices)==True:
    f = open(filePath_forecast_matrices,"rb")
    tau0 = pickle.load(f)
    G_tau_3d = pickle.load(f)
    Exp_StationErr = pickle.load(f)
    f.close()
else: 
    print('!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--')
    print('   WARNING: forecast matrix data file does not exist! Cannot continue.')
    print('!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--')
    sys.exit()
    
# Define size of data array 
sizes = np.shape(data)
nDat  = sizes[0]
nT    = sizes[1]

#----------------------------------------------------------
# Define weights on moisture observations for G_tau matrix
#----------------------------------------------------------

wgtMatrix = np.zeros([nDat,nDat])

nDat2 = int(nDat/2)
for i in range(nDat2):             # For the moisture norm, we only want to consider the bottom half of G. 
    wgtMatrix[i+nDat2,i+nDat2]=1   # So the weight matrix should be 1s on diagonal in bottom half

#--------------------------------------------------------------
# Number of lags determined in train_lim.py G_tau output data
#--------------------------------------------------------------
Gsize = np.shape(G_tau_3d)
n_days = Gsize[0]

gamma1  = np.full(n_days-1,np.nan)  #Empty array to store leading eigenvalue in 

#Check G(tau) for each value for each lag tau given tau0. Start on lag = 1 
for iLag in range(1,n_days):
    # Pull G_tau at lag=iLag from input data
    G_tau = G_tau_3d[iLag,:,:]

    # ---- EOF analysis of GtG ---- #
#    GTG = np.dot(np.transpose(G_tau),G_tau)  # Field norm
    GTG = np.dot(np.transpose(G_tau),np.dot(wgtMatrix, G_tau))  # Moisture norm

    eigVal_GTG, eigVec_GTG = LA.eig(GTG)

    # Sort the eigenvalues in order 
    iSort_GTG  = eigVal_GTG.argsort()[::-1]   
    eigVal_GTG = eigVal_GTG[iSort_GTG]
    eigVec_GTG = eigVec_GTG[:,iSort_GTG]

    # Save first eigen value for this Tau (the maximum growing mode)
    gamma1[iLag-1] = np.log(eigVal_GTG[0].real)

# Insert a value of zero for day zero (initial conditions = no amplification)
gamma_all = np.insert(gamma1, 0, 0)

#----------------------------------------------------------------------
# Define the lag associated with the maximum amplification (Tau_peak)
#---------------------------------------------------------------------
Tau_peak = int(np.where(gamma_all==np.max(gamma_all))[0])

#------------------------------------------------------------------------
# Determine optimal structure of moisture norm using lag with peak growth
#------------------------------------------------------------------------

#Define Green's function at the peak of the max amplification growth
G_tau = G_tau_3d[Tau_peak,:,:]
# Define Green's function for the moisture norm. nDat2 is half of nDat.
# lower half of matrix contains moisture data
G_m = G_tau[nDat2:nDat,:]

# EOF analysis of moisture data
GTG = np.dot(np.transpose(G_m),G_m)
eigVal_GTG, eigVec_GTG = LA.eig(GTG)

# EOF analysis of field data
#GTG = np.dot(np.transpose(G_tau),G_tau)
#eigVal_GTG, eigVec_GTG = LA.eig(GTG)

#Sort the eigenvalues in order 
iSort_GTG  = eigVal_GTG.argsort()[::-1]   
eigVal_GTG = eigVal_GTG[iSort_GTG] 
eigVec_GTG = eigVec_GTG[:,iSort_GTG]

#The optimal structure is the first eigenVector of the matrix GtG. 
optimStruct = eigVec_GTG[:,0].real

#Get the forecast by projecting G(tau) onto that optimal structure 
#forecast = np.dot(G_tau.real,optimStruct)

#---------------------------------------------------------------------------
# Two methods to compute histograms used to determine cut-off of top tercile
# Only using correlation method but leaving projection method in code
#---------------------------------------------------------------------------

## Projection method
#optProj   = np.full([nT], np.nan)
#for iT in range(nT):
    ##For matrix multiplication with nan, take nansum after *
#    optProj[iT]   = np.nansum(np.transpose(optimStruct) * data[:,iT])  

pcorr = np.full([nT], np.nan)
valid = np.full([nT], True)
valid_stations = np.full([nT], np.nan)
for iT in range(nT):
    valid = np.where(np.isfinite(data[:,iT]))[0]
    valid_stations[iT] = len(valid)
    if len(valid) > nDat/2:
        R = np.corrcoef(optimStruct[valid],data[valid,iT])
        pcorr[iT] = R[0,1]

#-------------------------------------------------------
# Define the cutoff for the top tercile for both methods
#-------------------------------------------------------
#projTrcile = np.nanpercentile(np.abs(optProj),66.66)   # Cutoff defined as top tercile here (can be changed)
pcorr_Trcile = np.nanpercentile(np.abs(pcorr),66.66)   # Cutoff defined as top tercile here (can be changed)

#-----------------------------------------------------------------------------------
# Output tau_peak, optimal structure, and tercile pattern correlation cut-off value
#-----------------------------------------------------------------------------------

f = open(dpath+"optimal_structure_peak.p","wb")
pickle.dump(Tau_peak, f)
pickle.dump(gamma_all, f)
pickle.dump(optimStruct, f)
f.close()

#f = open(dpath+"optimal_projection.p","wb")                              
#pickle.dump(projTrcile, f)
#pickle.dump(optProj, f)
#f.close()

f = open(dpath+"optimal_projection_pcorr.p","wb")
pickle.dump(pcorr_Trcile, f)
pickle.dump(pcorr,f)
f.close()
