#!/usr/bin/python
#
# NAME: verify_zscores.py
#
# PURPOSE: Reads zscore data created from test case soil moisture and
#          temperature data and compares with z-score data
#          provided with the model code
#
# INPUT DATA:
#             z-score data sets generated locally with supplied soil
#                  moisture and temperature data
#             z-score data provided to verify z-score data
#
# OUTPUT DATA: 
#              Correlation and max/min differences between z-score
#              data sets provide to standard output.
#
# NOTES:
#          Correlations at each station should be 1.
#          max/min differences should be zero.
#
# CREATION HISTORY:  Darren Jackson, CIRES/ESRL/PSL, April 2022
#----------------------------------------------------------------------

import os, sys
import numpy as np 
import pickle 

#--------------------------------------------------
# Input soil moisture and temperature zscore data 
#--------------------------------------------------

ipath = '../data/training_data/'
vpath = '../data/verify_data/'
filePath_zscores  = ipath+"zScores_residST_SM.p"
filePath_zscores_verify  = vpath+"zScores_residST_SM_filled_master.p"

if os.path.isfile(filePath_zscores)==True:
    data = pickle.load( open( filePath_zscores, "rb" ) )
else: 
   sys.exit('File = '+filePath_zscores+' does not exit. Program terminated.')

sizes = np.shape(data)
nDat = sizes[0]
nT = sizes[1]

if os.path.isfile(filePath_zscores_verify)==True:
    vdata = pickle.load( open( filePath_zscores_verify, "rb" ) )
else: 
   sys.exit('File = '+filePath_zscores_verify+' does not exit. Program terminated.')

#---------------------------------------------------------------------
# Generate statistics comparing zscore data against reference data
#---------------------------------------------------------------------

corr = np.full([nDat],np.nan)
for iDat in range(nDat):
    valid = np.where(np.isfinite(data[iDat,:]) & np.isfinite(vdata[iDat,:]))[0]
    R = np.corrcoef(data[iDat,valid],vdata[iDat,valid])
    corr[iDat] = R[0,1]
    
print('Correlation coefficient of z-scores at each station = ',corr)
diff = data - vdata
mx = np.nanmax(diff,axis=1)
mn = np.nanmin(diff,axis=1)
print('Maximum z-score difference at each station = ',mx)
print('Minimum z-score difference at each station = ',mn)
